@@ -13,13 +13,27 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from collections import namedtuple
+import logging
+import uuid
 import xml.etree.ElementTree as ET
 
-from collections import namedtuple
 from sushy_tools.emulator.drivers.base import AbstractDriver
-from sushy_tools.error import FishyError
+from sushy_tools.emulator.drivers import memoize
+from sushy_tools import error
+
+try:
+    import libvirt
+
+except ImportError:
+    libvirt = None
+
+
+is_loaded = bool(libvirt)
+
+
+logger = logging.getLogger(__name__)
 
-import libvirt
 
 BiosProcessResult = namedtuple('BiosProcessResult',
                                ['tree',
@@ -44,7 +58,7 @@ class libvirt_open(object):
         except libvirt.libvirtError as e:
             msg = ('Error when connecting to the libvirt URI "%(uri)s": '
                    '%(error)s' % {'uri': self._uri, 'error': e})
-            raise FishyError(msg)
+            raise error.FishyError(msg)
 
     def __exit__(self, type, value, traceback):
         self._conn.close()
@@ -67,18 +81,61 @@ class LibvirtDriver(AbstractDriver):
 
     BOOT_MODE_MAP = {
         'Legacy': 'rom',
-        'Uefi': 'pflash',
+        'Uefi': 'pflash'
     }
 
     BOOT_MODE_MAP_REV = {v: k for k, v in BOOT_MODE_MAP.items()}
 
+    BOOT_LOADER_MAP = {
+        'Uefi': {
+            'x86_64': '/usr/share/OVMF/OVMF_CODE.fd',
+            'aarch64': '/usr/share/AAVMF/AAVMF_CODE.fd'
+        },
+        'Legacy': {
+            'x86_64': None,
+            'aarch64': None
+        }
+
+    }
+
     DEFAULT_BIOS_ATTRIBUTES = {"BootMode": "Uefi",
                                "EmbeddedSata": "Raid",
                                "NicBoot1": "NetworkBoot",
                                "ProcTurboMode": "Enabled"}
 
-    def __init__(self, uri=None):
-        self._uri = uri or self.LIBVIRT_URI
+    @classmethod
+    def initialize(cls, config, uri=None):
+        cls._config = config
+        cls._uri = uri or cls.LIBVIRT_URI
+        cls.BOOT_LOADER_MAP = cls._config.get(
+            'SUSHY_EMULATOR_BOOT_LOADER_MAP', cls.BOOT_LOADER_MAP)
+        cls.KNOWN_BOOT_LOADERS = set(y for x in cls.BOOT_LOADER_MAP.values()
+                                     for y in x.values())
+        return cls
+
+    @memoize.memoize()
+    def _get_domain(self, identity, readonly=False):
+        with libvirt_open(self._uri, readonly=readonly) as conn:
+            try:
+                uu_identity = uuid.UUID(identity)
+
+                return conn.lookupByUUID(uu_identity.bytes)
+
+            except (ValueError, libvirt.libvirtError):
+                try:
+                    domain = conn.lookupByName(identity)
+
+                except libvirt.libvirtError as ex:
+                    msg = ('Error finding domain by name/UUID "%(identity)s" '
+                           'at libvirt URI %(uri)s": %(err)s' %
+                           {'identity': identity,
+                            'uri': self._uri, 'err': ex})
+
+                    logger.debug(msg)
+
+                    raise error.FishyError(msg)
+
+            raise error.AliasAccessError(domain.UUIDString())
 
     @property
     def driver(self):
@@ -92,10 +149,11 @@ class LibvirtDriver(AbstractDriver):
     def systems(self):
         """Return available computer systems
 
-        :returns: list of computer systems names.
+        :returns: list of UUIDs representing the systems
         """
         with libvirt_open(self._uri, readonly=True) as conn:
-            return conn.listDefinedDomains()
+            return [conn.lookupByName(domain).UUIDString()
+                    for domain in conn.listDefinedDomains()]
 
     def uuid(self, identity):
         """Get computer system UUID
@@ -103,13 +161,22 @@ class LibvirtDriver(AbstractDriver):
         The universal unique identifier (UUID) for this system. Can be used
         in place of system name if there are duplicates.
 
-        :param identity: libvirt domain name or ID
+        :param identity: libvirt domain name or UUID
 
         :returns: computer system UUID
         """
-        with libvirt_open(self._uri, readonly=True) as conn:
-            domain = conn.lookupByName(identity)
-            return domain.UUIDString()
+        domain = self._get_domain(identity, readonly=True)
+        return domain.UUIDString()
+
+    def name(self, identity):
+        """Get computer system name by name
+
+        :param identity: libvirt domain name or UUID
+
+        :returns: computer system name
+        """
+        domain = self._get_domain(identity, readonly=True)
+        return domain.name()
 
     def get_power_state(self, identity):
         """Get computer system power state
@@ -119,10 +186,8 @@ class LibvirtDriver(AbstractDriver):
         :returns: current power state as *On* or *Off* `str` or `None`
             if power state can't be determined
         """
-        with libvirt_open(self._uri, readonly=True) as conn:
-
-            domain = conn.lookupByName(identity)
-            return 'On' if domain.isActive() else 'Off'
+        domain = self._get_domain(identity, readonly=True)
+        return 'On' if domain.isActive() else 'Off'
 
     def set_power_state(self, identity, state):
         """Set computer system power state
@@ -133,37 +198,35 @@ class LibvirtDriver(AbstractDriver):
             Valid values  are: *On*, *ForceOn*, *ForceOff*, *GracefulShutdown*,
             *GracefulRestart*, *ForceRestart*, *Nmi*.
 
-        :raises: `FishyError` if power state can't be set
+        :raises: `error.FishyError` if power state can't be set
         """
-        with libvirt_open(self._uri) as conn:
+        domain = self._get_domain(identity)
 
-            domain = conn.lookupByName(identity)
+        try:
+            if state in ('On', 'ForceOn'):
+                if not domain.isActive():
+                    domain.create()
+            elif state == 'ForceOff':
+                if domain.isActive():
+                    domain.destroy()
+            elif state == 'GracefulShutdown':
+                if domain.isActive():
+                    domain.shutdown()
+            elif state == 'GracefulRestart':
+                if domain.isActive():
+                    domain.reboot()
+            elif state == 'ForceRestart':
+                if domain.isActive():
+                    domain.reset()
+            elif state == 'Nmi':
+                if domain.isActive():
+                    domain.injectNMI()
 
-            try:
-                if state in ('On', 'ForceOn'):
-                    if not domain.isActive():
-                        domain.create()
-                elif state == 'ForceOff':
-                    if domain.isActive():
-                        domain.destroy()
-                elif state == 'GracefulShutdown':
-                    if domain.isActive():
-                        domain.shutdown()
-                elif state == 'GracefulRestart':
-                    if domain.isActive():
-                        domain.reboot()
-                elif state == 'ForceRestart':
-                    if domain.isActive():
-                        domain.reset()
-                elif state == 'Nmi':
-                    if domain.isActive():
-                        domain.injectNMI()
+        except libvirt.libvirtError as e:
+            msg = ('Error changing power state at libvirt URI "%(uri)s": '
+                   '%(error)s' % {'uri': self._uri, 'error': e})
 
-            except libvirt.libvirtError as e:
-                msg = ('Error changing power state at libvirt URI "%(uri)s": '
-                       '%(error)s' % {'uri': self._uri, 'error': e})
-
-                raise FishyError(msg)
+            raise error.FishyError(msg)
 
     def get_boot_device(self, identity):
         """Get computer system boot device name
@@ -173,19 +236,17 @@ class LibvirtDriver(AbstractDriver):
         :returns: boot device name as `str` or `None` if device name
             can't be determined
         """
-        with libvirt_open(self._uri, readonly=True) as conn:
+        domain = self._get_domain(identity, readonly=True)
 
-            domain = conn.lookupByName(identity)
+        tree = ET.fromstring(domain.XMLDesc())
 
-            tree = ET.fromstring(domain.XMLDesc())
+        boot_element = tree.find('.//boot')
+        if boot_element is not None:
+            boot_source_target = (
+                self.BOOT_DEVICE_MAP_REV.get(boot_element.get('dev'))
+            )
 
-            boot_element = tree.find('.//boot')
-            if boot_element is not None:
-                boot_source_target = (
-                    self.BOOT_DEVICE_MAP_REV.get(boot_element.get('dev'))
-                )
-
-                return boot_source_target
+            return boot_source_target
 
     def set_boot_device(self, identity, boot_source):
         """Get/Set computer system boot device name
@@ -195,41 +256,40 @@ class LibvirtDriver(AbstractDriver):
             change on the system. If not specified, current boot device is
             returned. Valid values are: *Pxe*, *Hdd*, *Cd*.
 
-        :raises: `FishyError` if boot device can't be set
+        :raises: `error.FishyError` if boot device can't be set
         """
-        with libvirt_open(self._uri) as conn:
+        domain = self._get_domain(identity)
 
-            domain = conn.lookupByName(identity)
+        # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
+        tree = ET.fromstring(domain.XMLDesc())
 
-            # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
-            tree = ET.fromstring(domain.XMLDesc())
+        try:
+            target = self.BOOT_DEVICE_MAP[boot_source]
 
-            try:
-                target = self.BOOT_DEVICE_MAP[boot_source]
+        except KeyError:
+            msg = ('Unknown power state requested: '
+                   '%(boot_source)s' % {'boot_source': boot_source})
 
-            except KeyError:
-                msg = ('Unknown power state requested: '
-                       '%(boot_source)s' % {'boot_source': boot_source})
+            raise error.FishyError(msg)
 
-                raise FishyError(msg)
+        for os_element in tree.findall('os'):
+            # Remove all "boot" elements
+            for boot_element in os_element.findall('boot'):
+                os_element.remove(boot_element)
 
-            for os_element in tree.findall('os'):
-                # Remove all "boot" elements
-                for boot_element in os_element.findall('boot'):
-                    os_element.remove(boot_element)
+            # Add a new boot element with the request boot device
+            boot_element = ET.SubElement(os_element, 'boot')
+            boot_element.set('dev', target)
 
-                # Add a new boot element with the request boot device
-                boot_element = ET.SubElement(os_element, 'boot')
-                boot_element.set('dev', target)
-
-            try:
+        try:
+            with libvirt_open(self._uri) as conn:
                 conn.defineXML(ET.tostring(tree).decode('utf-8'))
 
-            except libvirt.libvirtError as e:
-                msg = ('Error changing boot device at libvirt URI "%(uri)s": '
-                       '%(error)s' % {'uri': self._uri, 'error': e})
+        except libvirt.libvirtError as e:
+            msg = ('Error changing boot device at libvirt URI "%(uri)s": '
+                   '%(error)s' % {'uri': self._uri, 'error': e})
 
-                raise FishyError(msg)
+            raise error.FishyError(msg)
 
     def get_boot_mode(self, identity):
         """Get computer system boot mode.
@@ -237,21 +297,19 @@ class LibvirtDriver(AbstractDriver):
         :returns: either *Uefi* or *Legacy* as `str` or `None` if
             current boot mode can't be determined
         """
-        with libvirt_open(self._uri, readonly=True) as conn:
+        domain = self._get_domain(identity, readonly=True)
 
-            domain = conn.lookupByName(identity)
+        # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
+        tree = ET.fromstring(domain.XMLDesc())
 
-            # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
-            tree = ET.fromstring(domain.XMLDesc())
+        loader_element = tree.find('.//loader')
 
-            loader_element = tree.find('.//loader')
+        if loader_element is not None:
+            boot_mode = (
+                self.BOOT_MODE_MAP_REV.get(loader_element.get('type'))
+            )
 
-            if loader_element is not None:
-                boot_mode = (
-                    self.BOOT_MODE_MAP_REV.get(loader_element.get('type'))
-                )
-
-                return boot_mode
+            return boot_mode
 
     def set_boot_mode(self, identity, boot_mode):
         """Set computer system boot mode.
@@ -260,41 +318,82 @@ class LibvirtDriver(AbstractDriver):
             change on the system. If not specified, current boot mode is
             returned. Valid values are: *Uefi*, *Legacy*.
 
-        :raises: `FishyError` if boot mode can't be set
+        :raises: `error.FishyError` if boot mode can't be set
         """
-        with libvirt_open(self._uri) as conn:
+        domain = self._get_domain(identity, readonly=True)
 
-            domain = conn.lookupByName(identity)
+        # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
+        tree = ET.fromstring(domain.XMLDesc())
 
-            # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
-            tree = ET.fromstring(domain.XMLDesc())
+        try:
+            loader_type = self.BOOT_MODE_MAP[boot_mode]
+
+        except KeyError:
+            msg = ('Unknown boot mode requested: '
+                   '%(boot_mode)s' % {'boot_mode': boot_mode})
+
+            raise error.FishyError(msg)
+
+        loaders = []
+
+        for os_element in tree.findall('os'):
+            type_element = os_element.find('type')
+            if type_element is None:
+                os_arch = None
+            else:
+                os_arch = type_element.get('arch')
 
             try:
-                target = self.BOOT_MODE_MAP[boot_mode]
+                loader_path = self.BOOT_DEVICE_MAP[boot_mode][os_arch]
 
             except KeyError:
-                msg = ('Unknown boot mode requested: '
-                       '%(boot_mode)s' % {'boot_mode': boot_mode})
+                logger.warning('Boot loader is not configured for '
+                               'boot mode %s and OS architecture %s. '
+                               'Assuming no specific boot loader.',
+                               boot_mode, os_arch)
+                loader_path = ''
 
-                raise FishyError(msg)
+            # NOTE(etingof): here we collect all tree nodes to be
+            # updated, but do not update them yet. The reason is to
+            # make sure that previously configured boot loaders are
+            # all present in our configuration, so we won't lose it
+            # when setting ours.
 
-            for os_element in tree.findall('os'):
-                # Remove all "boot" elements
-                for loader_element in os_element.findall('loader'):
-                    os_element.remove(loader_element)
+            for loader_element in os_element.findall('loader'):
+                if loader_element.text not in self.KNOWN_BOOT_LOADERS:
+                    msg = ('Unknown boot loader path "%(path)s" in domain '
+                           '"%(identity)s" configuration encountered while '
+                           'setting boot mode "%(mode)s", system architecture '
+                           '"%(arch)s". Consider adding this loader path to '
+                           'emulator config.' % {'identity': identity,
+                                                 'mode': boot_mode,
+                                                 'arch': os_arch,
+                                                 'path': loader_element.text})
+                    raise error.FishyError(msg)
 
-                # Add a new loader element with the request boot mode
-                loader_element = ET.SubElement(os_element, 'loader')
-                loader_element.set('type', target)
+                loaders.append((loader_element, loader_path))
+
+        if not loaders:
+            msg = ('Can\'t set boot mode because no "os" elements are present '
+                   'in domain "%(identity)s" or not a single "os" element has '
+                   '"loader" configured.' % {'identity': identity})
+            raise error.FishyError(msg)
+
+        for loader_element, loader_path in loaders:
+            loader_element.set('type', loader_type)
+            loader_element.text = loader_path
+
+        with libvirt_open(self._uri) as conn:
 
             try:
-                self._conn.defineXML(ET.tostring(tree).decode('utf-8'))
+                conn.defineXML(ET.tostring(tree).decode('utf-8'))
 
             except libvirt.libvirtError as e:
-                msg = ('Error changing boot mode at libvirt URI "%(uri)s": '
-                       '%(error)s' % {'uri': self._uri, 'error': e})
+                msg = ('Error changing boot mode at libvirt URI '
+                       '"%(uri)s": %(error)s' % {'uri': self._uri,
+                                                 'error': e})
 
-                raise FishyError(msg)
+                raise error.FishyError(msg)
 
     def get_total_memory(self, identity):
         """Get computer system total memory
@@ -304,9 +403,8 @@ class LibvirtDriver(AbstractDriver):
         :returns: available RAM in GiB as `int` or `None` if total memory
             count can't be determined
         """
-        with libvirt_open(self._uri, readonly=True) as conn:
-            domain = conn.lookupByName(identity)
-            return int(domain.maxMemory() / 1024 / 1024)
+        domain = self._get_domain(identity, readonly=True)
+        return int(domain.maxMemory() / 1024 / 1024)
 
     def get_total_cpus(self, identity):
         """Get computer system total count of available CPUs
@@ -316,25 +414,23 @@ class LibvirtDriver(AbstractDriver):
         :returns: available CPU count as `int` or `None` if CPU count
             can't be determined
         """
-        with libvirt_open(self._uri, readonly=True) as conn:
+        domain = self._get_domain(identity, readonly=True)
 
-            domain = conn.lookupByName(identity)
+        tree = ET.fromstring(domain.XMLDesc())
 
-            tree = ET.fromstring(domain.XMLDesc())
+        total_cpus = 0
 
-            total_cpus = 0
+        if domain.isActive():
+            total_cpus = domain.maxVcpus()
 
-            if domain.isActive():
-                total_cpus = domain.maxVcpus()
+        # If we can't get it from maxVcpus() try to find it by
+        # inspecting the domain XML
+        if total_cpus <= 0:
+            vcpu_element = tree.find('.//vcpu')
+            if vcpu_element is not None:
+                total_cpus = int(vcpu_element.text)
 
-            # If we can't get it from maxVcpus() try to find it by
-            # inspecting the domain XML
-            if total_cpus <= 0:
-                vcpu_element = tree.find('.//vcpu')
-                if vcpu_element is not None:
-                    total_cpus = int(vcpu_element.text)
-
-            return total_cpus or None
+        return total_cpus or None
 
     def _process_bios_attributes(self,
                                  domain_xml,
@@ -415,23 +511,27 @@ class LibvirtDriver(AbstractDriver):
 
         :returns: New or existing dict of BIOS attributes
 
-        :raises: `FishyError` if BIOS attributes cannot be saved
+        :raises: `error.FishyError` if BIOS attributes cannot be saved
         """
-        with libvirt_open(self._uri) as conn:
-            libvirt_domain = conn.lookupByName(identity)
-            result = self._process_bios_attributes(libvirt_domain.XMLDesc(),
-                                                   bios_attributes,
-                                                   update_existing_attributes)
+        domain = self._get_domain(identity)
 
-            if result.attributes_written:
-                try:
+        result = self._process_bios_attributes(domain.XMLDesc(),
+                                               bios_attributes,
+                                               update_existing_attributes)
+
+        if result.attributes_written:
+
+            try:
+                with libvirt_open(self._uri) as conn:
                     conn.defineXML(ET.tostring(result.tree).decode('utf-8'))
-                except libvirt.libvirtError as e:
-                    msg = ('Error updating BIOS attributes'
-                           ' at libvirt URI "%(uri)s": '
-                           '%(error)s' % {'uri': self._uri, 'error': e})
-                    raise FishyError(msg)
-            return result.bios_attributes
+
+            except libvirt.libvirtError as e:
+                msg = ('Error updating BIOS attributes'
+                       ' at libvirt URI "%(uri)s": '
+                       '%(error)s' % {'uri': self._uri, 'error': e})
+                raise error.FishyError(msg)
+
+        return result.bios_attributes
 
     def get_bios(self, identity):
         """Get BIOS section
@@ -472,3 +572,18 @@ class LibvirtDriver(AbstractDriver):
         """
         self._process_bios(identity, self.DEFAULT_BIOS_ATTRIBUTES,
                            update_existing_attributes=True)
+
+    def get_nics(self, identity):
+        """Get list of network interfaces and their MAC addresses
+
+        Use MAC address as network interface's id
+
+        :param identity: libvirt domain name or ID
+
+        :returns: list of network interfaces dict with their attributes
+        """
+        domain = self._get_domain(identity, readonly=True)
+        tree = ET.fromstring(domain.XMLDesc())
+        return [{'id': iface.get('address'), 'mac': iface.get('address')}
+                for iface in tree.findall(
+                ".//devices/interface[@type='network']/mac")]
